# Given a response Y and predictors X, fit a multivariate linear regression
# and return its parameters
estimate_model <- function(Y, X, estimateCov = F, t=NULL) {
  Y <- as.matrix(Y)
  X <- as.matrix(X)
  d <- ncol(Y)
  k <- ncol(X)
  n <- nrow(Y)
  
  if (nrow(Y) > (k + 1)) {
    # compare means of Y over the years
    
    doit <- function() {
      Bhat <- matrix(0, k, d)
      resids <- matrix(0, n, d)
      for (i in 1:d) {
        fm1 <- lm(Y[, i] ~ -1 + X)
        resids[, i] <- fm1$residuals
        Bhat[, i] <- fm1$coefficients
      }
      
      Sigmahat <- cov(resids)
      
      par_vals <- list(Sigmahat = Sigmahat, Bhat = Bhat)
      
      return(par_vals)
    }
    par_vals_try <- try(
      doit(),
      silent = T
    )
    if (inherits(par_vals_try, "try-error")) {
      par_vals <- list(Sigmahat = matrix(NA, d, d), Bhat = matrix(NA, k, d))
    } else {
      if (any(is.na(c(unlist(par_vals_try))))) {
        par_vals <- list(Sigmahat = matrix(NA, d, d), Bhat = matrix(NA, k, d))
      }
      par_vals <- par_vals_try
    }
  } else {
    par_vals <- list(Sigmahat = matrix(NA, d, d), Bhat = matrix(NA, k, d))
  }
  
  return(list(prior_vals = par_vals))
}



#' get_priors: calculate hyperparameters from estimated mean and covariance
#'
#' @param data_list
#' @param dpriors
#' @param k
#' @param d
#'
#' @return
#' @export
#'
#' @examples
get_priors <- function(data_list, dpriors, k = 4, d = 2) {
  Sigmas <- lapply(data_list, function(x) {
    if (any(is.na(x$prior_vals$Sigmahat))) {
      return(matrix(NA, d, d))
    } else {
      invertable <- try(solve(x$prior_vals$Sigmahat), silent = T)
      if (inherits(invertable, "try-error")) {
        return(matrix(NA, d, d))
      } else {
        return(x$prior_vals$Sigmahat)
      }
    }
  })
  Betas <- lapply(data_list, function(x) {
    x$prior_vals$Bhat
  })
  n <- sum(sapply(Sigmas, function(x) {
    !any(is.na(x))
  }))
  
  logdetV <- sum(sapply(Sigmas, function(x) {
    if (any(is.na(x))) {
      return(NA)
    } else {
      return(log(det(x)))
    }
  }), na.rm = T)
  
  meanV <- matrix(rowMeans(do.call(cbind, lapply(Sigmas, function(x) {
    c(x)
  })), na.rm = T), d, d)
  
  
  
  f_nu <- function(nu, d, n, logdetV, meanV) {
    val1 <- (n / 2) * log(det((nu - d - 1) * meanV))
    val2 <- -n * d * log(2) / 2
    val3 <- -0.5 * logdetV
    val4 <- -(n / 2) * sum(sapply(1:d, function(j) {
      digamma(nu / 2 + (1 - j) / 2)
    }))
    val1 + val2 + val3 + val4
  }
  
  
  doit <- function() {
    root_result <- uniroot(f_nu, interval = c(d + 1 + .001, 200), d = d, n = n, logdetV = logdetV, meanV = meanV)
    
    nuhat <- root_result$root
    
    Vhat <- meanV * (nuhat - d - 1)
    # B <- t(sapply(Betas,function(x){c(x)}))
    B <- do.call(rbind, lapply(Betas, function(x) {
      c(x)
    }))
    Bhat <- matrix(colMeans(B, na.rm = T), k, d)
    
    LambdaInvcomponents <- t(mapply(function(Bi, Sigmai, Bhat) {
      val1 <- Bi - Bhat
      val1 %*% solve(Sigmai) %*% t(val1)
    }, Bi = Betas, Sigmai = Sigmas, MoreArgs = list(Bhat = Bhat)))
    LambdaInvcomponents[rowSums(LambdaInvcomponents > 1000, na.rm = T) != 0, ] <- NA
    LambdaInvhat <- (1 / d) * matrix(colMeans(LambdaInvcomponents, na.rm = T), k, k)
    
    Lambdahat <- solve(LambdaInvhat)
    
    return(list(B = Bhat, V = Vhat, Lambda = Lambdahat, nu = nuhat))
  }
  par_vals_try <- try(
    doit(),
    silent = T
  )
  if (inherits(par_vals_try, "try-error")) {
    B <- do.call(rbind, lapply(Betas, function(x) {
      c(x)
    }))
    Bhat <- matrix(colMeans(B, na.rm = T), k, d)
    nuhat <- d + 1 + .01
    Vhat <- meanV * (nuhat - d - 1)
    #
    # LambdaInv <- solve(dpriors$Lambda)
    # LambdahatInv <- matrix(colMeans(do.call(rbind,lapply(1:d,function(x){
    #   c(LambdaInv*(dpriors$V[x,x]/(dpriors$nu-d-1))/(Vhat[x,x]/(nuhat-d-1)))
    # }))),k,k)
    # Lambdahat <- solve(LambdahatInv)
    #
    par_vals <- list(B = Bhat, V = Vhat, Lambda = matrix(NA, k, k), nu = nuhat)
  } else {
    par_vals <- par_vals_try
  }
  
  return(par_vals)
}
#' Calculate prior parameters
#'
#' @param signal_cube array of dimension pixels x time x signal.
#' @param change_date_p The last date of the stable period
#' @param ptype Can take the following values:
#'             "default" or 0 returns a default prior specification, not based on data
#'             "all" or 1 returns on set of priors based on a sample of \code{nsample} pixels
#'             "selection" or 5 returns priors based on a selection of pixels in \code{selection}
#' @param selection when \code{ptype} is "selection", \code{selection} is a list containing lists of pixel numbers from which to learn priors
#' @param lambda_default if T, lambda will revert to the default (recommended). If F, lambda will be estimated from data
#' @param model_pars list of model parameters
#' @param ...
#' @param nsample number of samples to take for each entry according to \code{ptype}. If nsample > number of pixels, all pixels will be used to calculate priors
#'
#' @return A list containing \code{priors} associated with the \code{ptype} and a lookup table \code{prior_lookup} specifying the point (column 1) and its location in the prior list (column 2)
#' @examples
calculate_priors <- function(signals, X, nsample = 10000, ...) {
  # get number of pixels
  npix <- dim(signals)[[1]]
  # get number of signals
  d <- dim(signals)[3]
  # get number of covariates
  k <- ncol(X)
  # make default priors
  default_priors <- list(B=matrix(0,ncol(signals),ncol(X)),
                         Lambda=diag(c(0.01,rep(0.1,ncol(X)-2),10)),
                         V = diag(1,d),
                         nu = d+.01)
  nsamp <- 500
  if(nsamp<npix){
    pixels_sample <- sample(1:npix,nsamp)
    
  }else{pixels_sample <- 1:npix}
  
  
  sample_data_list <- lapply(1:length(pixels_sample), FUN = function(w) {
    pixel_num <- pixels_sample[w]
    Yi <- signals[pixel_num,,]
    datapts <- estimate_model(Yi,X)
    return(datapts)
  })
  
  # # calculate priors based on training data for each subregion
  priors_list <-  list(get_priors(data_list = sample_data_list, d = d, dpriors = default_priors, k = k))
  
  ### verify that priors are not NA and fill in with default wherever necessary
  priors <- lapply(priors_list, function(x) {
    if (any(is.na(x$B))) {
      B <- default_priors$B
    } else {
      B <- x$B
    }
    if (any(is.na(x$V))) {
      V <- default_priors$V
    } else {
      V <- x$V
    }
    if (any(is.na(x$nu))) {
      nu <- default_priors$nu
    } else {
      nu <- x$nu
    }
    if (any(is.na(x$Lambda))) {
      Lambda <- default_priors$Lambda
    } else {
      Lambda <- x$Lambda
    }
    return(list(B = B, V = V, nu = nu, Lambda = Lambda))
  })
  
  return(list(priors = priors))
}

set.seed(357876)
# 
# # simulate stable data
# dat <- simulate_data(J=J,stable=F)[,,1:d,]
# 
# # decompose image
# dat_w <- get_decomps(array(dat[,,1:d,],dim=c(256,256,d,80)))

#

## wavelet priors
xvals <- 1:256
yvals <- 1:256
d <- 2
prior_list <- list()
for(j in Jvals){
  # extract data_cube sample
  LL <- paste("LL",J,sep="")
  LH <- paste("LH", j, sep = "")
  HL <- paste("HL", j, sep = "")
  HH <- paste("HH", j, sep = "")
  if(j==J){
    mvec <- c(LH,HL,HH,LL)
    
  }else{mvec <- c(LH, HL, HH)}
  for(m in seq_along(mvec)){
    data_cube_mj <- array(dat_w[[m]],dim=c(prod(dim(dat_w[[m]])[1:2]),length(tvals),d))
    priors <- calculate_priors(data_cube_mj,X)
    newnames <- c(names(prior_list),mvec[m])
    prior_list[[mvec[m]]] <- priors
    names(prior_list) <- newnames
  }
}
save(prior_list,file=file_priors)


## classic priors
data_cube_i <- array(dat,dim=c(prod(dim(dat)[1:2]),length(tvals),d))
priors <- calculate_priors(data_cube_i,X)
prior_list <- list(priors)
save(prior_list,file=file_priors_classic)
