## ---------------------------
##
## Script name: sim_example
##
## Purpose of script: Runs all variants of pixelwise/MR roboBayes
## described in the paper Multiresolution Broad Area Search: Monitoring Spatial
## Characteristics of Gapless Remote Sensing Data
##
## Author: Laura Wendelberger
##
## Date Created: 2022-03-24
##
## Email: ljwendel@ncsu.edu
##
## ---------------------------
##
## Notes:
##   
##
## ---------------------------
library(spectralGP)
library(waveslim)
library(ncdf4)
library(roboBayes)
library(data.table)
library(terra)
library(fields)
library(foreach)
library(parallel)
library(future.apply)
library(optparse)
n_cores <- 16


option_list <- list(
  make_option(c("-t","--tag"),
              type = "character",
              default = "_sim12"
  ),
  make_option(c("-l","--Lambda"),
              type = "numeric",
              default = 0
  ),
  make_option(c("-v","--vscale"),
              type = "numeric",
              default = 1
  ),
  make_option(c("-n","--nu"),
              type = "numeric",
              default = 0
  ),
  make_option(c("-s","--Lsearch"),
              type="integer",
              default=30)
)



opt_parser <- OptionParser(option_list = option_list)
opt <- parse_args(opt_parser)

tag <- opt$tag
vscale <- opt$vscale
Lambda <- opt$Lambda
nu <- opt$nu
opt$seas <- 1
seas <- 1
Lsearch <- opt$Lsearch

priors_adj <- list(vscale=vscale,nu=nu,Lambda=Lambda)

xdim <- 256
xvals <- 1:xdim
ydim <- 256
yvals <- 1:ydim
J <- 8
d <- 2
Lsearch <- opt$Lsearch

Jvals <- 1:8

source("functions/funs_simple.R")

# create the true detections
truth_mask <- simulate_data(J=J,d=d,ti=80, getTruth=T,Lsearch=Lsearch+3)
truth_mask_condensed <- apply(truth_mask,c(1,2),function(x){
  any(x>=1)
})

# set up covariates
tvals <- 1:80
X <- cbind(1,sin(2*pi*tvals/365),cos(2*pi*tvals/365),
           tvals/100)
tmax <- max(tvals)

# load or create priors
file_priors <- paste("interim_files/priors_sim.RData",sep="")
file_priors_classic <- paste("interim_files/priors_classic_sim.RData",sep="")
#source("scripts/get_sim_priors.R")
# load priors
load(file_priors)
priors_MRi <- prior_list
load(file_priors_classic)
priors_classici <- prior_list


# adjust priors per above adjustments
priors_MR <- lapply(priors_MRi,function(active_priors){
  priors_adj_j <- active_priors
  if(priors_adj$Lambda==0){
    priors_adj_j$priors[[1]]$Lambda <- diag(c(1,rep(0.1,2*seas),10))
  }else{
    priors_adj_j$priors[[1]]$Lambda <- priors_adj$Lambda*diag(c(diag(active_priors$priors[[1]]$Lambda)[1:(seas*2+1)],1e2))
  }
  priors_adj_j$priors[[1]]$V <- priors_adj$vscale*active_priors$priors[[1]]$V
  if(priors_adj$nu==0){
    priors_adj_j$priors[[1]]$nu <-  active_priors$priors[[1]]$nu
  }else{ priors_adj_j$priors[[1]]$nu <- priors_adj$nu}
  return(priors_adj_j)
})
priors_classic <- lapply(priors_classici,function(active_priors){
  priors_adj_j <- active_priors
  if(priors_adj$Lambda==0){
    priors_adj_j$priors[[1]]$Lambda <- diag(c(1,rep(0.1,2*seas),10))
  }else{
    priors_adj_j$priors[[1]]$Lambda <- priors_adj$Lambda*diag(c(diag(active_priors$priors[[1]]$Lambda)[1:(seas*2+1)],1e2))
  }
  priors_adj_j$priors[[1]]$V <- priors_adj$vscale*active_priors$priors[[1]]$V
  if(priors_adj$nu==0){
    priors_adj_j$priors[[1]]$nu <-  active_priors$priors[[1]]$nu
  }else{ priors_adj_j$priors[[1]]$nu <- priors_adj$nu}
  return(priors_adj_j)
})

# wavelet roboBayes
mvec <- c("LH","HL")

set.seed(26438)
# simulate data
dat <- simulate_data(J=J,stable=F)[,,1:d,]

# decompose image data with wavelet decomposition
dat_w <- get_decomps(array(dat[,,1:d,],dim=c(256,256,d,80)))

# monitor wavelet coefficients
Jvals <- 1:5 # levels 1-5
prob_vals <- c()
prob_array <- array(0,dim=c(2^(J-min(Jvals)),2^(J-min(Jvals)),tmax,length(Jvals)*length(mvec)))

for(m in seq_along(mvec)){
  for(j in Jvals){
    q_step <- min(2^(J-j),32)
    q_start <- expand.grid(seq(1,2^(J-j),q_step),seq(1,2^(J-j),q_step))
    
    #valmat <- array(0,dim=c(2^(J-j),2^(J-j),tmax))
    valmat <- array(0,dim=c(q_step,q_step,tmax))
    signal <- dat_w[[paste(mvec[m],j,sep="")]]
    prs <- priors_MR[[paste(mvec[m],j,sep="")]]$priors[[1]]
    # run roboBayes in parallel
    cl <- parallel::makeCluster(n_cores)
    doParallel::registerDoParallel(cl)
    #for(n in (1:(2^(J-j)))){
    result <- foreach(n = 1:nrow(q_start),
                      .packages = c("roboBayes","ncdf4","waveslim"),
                      .errorhandling = "stop") %dopar% {
                        for(qi in 1:q_step){
                          for(qj in 1:q_step){
                            mod <- roboBayes(datapts=t(signal[q_start[n,1]+qi-1,q_start[n,2]+qj-1,,]),
                                             covariates=X,kt=1,
                                             lambda=1e8,
                                             par_inits=prs,
                                             Lm=7,getOutliers=T,
                                             Lsearch=Lsearch,
                                             outlier_mean=rep(0,d),
                                             outlier_var = 2*(prs$nu-d-1)*diag(c(diag(prs$V)/prs$Lambda[1,1]),d))
                            valmat[qi,qj,] <- mod$lastLs
                          }
                        }
                        return(valmat)
                      }
    parallel::stopCluster(cl)
    val.array <- array(0,dim=c(2^(J-j),2^(J-j),length(tvals)))
    for(n in 1:nrow(q_start)){
      val.array[q_start[n,1]:(q_start[n,1]+q_step-1),q_start[n,2]:(q_start[n,2]+q_step-1),] <- result[[n]]
    }
    #val.array <- array(result,c(length(tvals),2^(J-j),2^(J-j)))
    #print(c(m,j,(m-1)*length(Jvals)+j-min(Jvals)+1))
    valmat.rast <- rast(val.array)
    if(j!=min(Jvals)){
      val <- terra::values(disagg(valmat.rast,fact=2^(j-min(Jvals))))
    }else{val <- terra::values(valmat.rast)}
    prob_array[,,,(m-1)*length(Jvals)+j-min(Jvals)+1] <- val
  }
}


# aggregate in different ways
set1 <- c(1:4,6:9)
set2 <- c(3:5,8:10)
# prob1
MR_roboBayes_res1 <- apply(prob_array[,,,set1],c(1,2,3),wavelet_to_realspace_probchange,dg=1)
# prob2
MR_roboBayes_res2 <- apply(prob_array[,,,set1],c(1,2,3),wavelet_to_realspace_probchange,dg=2)

# prob1 subset
MR_roboBayes_res1_2 <- apply(prob_array[,,,set2],c(1,2,3),wavelet_to_realspace_probchange,dg=1)
# prob2 subset
MR_roboBayes_res2_2 <- apply(prob_array[,,,set2],c(1,2,3),wavelet_to_realspace_probchange,dg=2)

# count wavelet components
MR_roboBayes_count <- apply(prob_array[,,,set1],c(1,2,3),count_wavelet_components,thresh=0.8)
MR_roboBayes_count_subset <- apply(prob_array[,,,set2],c(1,2,3),count_wavelet_components,thresh=0.8)



# run roboBayes on pixels
pixel_probs <- array(0,dim=c(2^J,2^J,tmax))
prs <- priors_classic[[1]]$priors[[1]]

q_step <- 32
q_start <- expand.grid(seq(1,2^J,q_step),seq(1,2^J,q_step))

#valmat <- array(0,dim=c(2^(J-j),2^(J-j),tmax))
valmat <- array(0,dim=c(q_step,q_step,tmax))
# run roboBayes in parallel
cl <- parallel::makeCluster(n_cores)
doParallel::registerDoParallel(cl)
#for(n in (1:(2^(J-j)))){
result <- foreach(n = 1:nrow(q_start),
                  .packages = c("roboBayes","ncdf4","waveslim"),
                  .errorhandling = "stop") %dopar% {
                    for(qi in 1:q_step){
                      for(qj in 1:q_step){
                        mod <- roboBayes(t(dat[q_start[n,1]+qi-1,q_start[n,2]+qj-1,1:d,]),
                                         par_inits=prs,
                                         covariates=X,kt=1,
                                         lambda=1e8,
                                         Lm=7,getOutliers=T,
                                         Lsearch=Lsearch,
                                         outlier_mean=rep(0,d),
                                         outlier_var = 2*(prs$nu-d-1)*diag(c(diag(prs$V)/prs$Lambda[1,1]),d))
                        
                        valmat[qi,qj,] <- mod$lastLs
                      }
                    }
                    return(valmat)
                  }
parallel::stopCluster(cl)
pixel_probs <- array(0,dim=c(2^(J),2^(J),length(tvals)))
for(n in 1:nrow(q_start)){
  pixel_probs[q_start[n,1]:(q_start[n,1]+q_step-1),q_start[n,2]:(q_start[n,2]+q_step-1),] <- result[[n]]
}




results_list <- list(MR_roboBayes_res1,MR_roboBayes_res2,
                     MR_roboBayes_res1_2,MR_roboBayes_res2_2,
                     MR_roboBayes_count, MR_roboBayes_count,
                     MR_roboBayes_count, MR_roboBayes_count,
                     MR_roboBayes_count, MR_roboBayes_count,
                     MR_roboBayes_count_subset, MR_roboBayes_count_subset,
                     MR_roboBayes_count_subset, MR_roboBayes_count_subset,
                     MR_roboBayes_count_subset, MR_roboBayes_count_subset,
                     aperm(pixel_probs,c(2,1,3)))

thresh_list <- c(rep(0.8,4),1:6,1:6,0.8)
poly_results <- list(1:length(thresh_list))
annots_flat <- terra::as.polygons((patches(rast(truth_mask_condensed),zeroAsNA=T)),values=T)
annots_flat$dates <- tchange

e <- ext(annots_flat)
e[1:4] <- c(0,256,0,256)
# # Get change polygons for each method
for(k in 1:length(results_list)){
  thresh <- thresh_list[k]
  post_probs <- results_list[[k]]
  post_change <- post_probs
  post_change[post_probs<thresh] <- 0
  post_change[post_probs>=thresh] <- 1
  post_change_first <- apply(post_change[,,18:80],c(1,2),function(x){which(x==1)[1]})
  
  post_change_condensed <- rast(t(apply(post_change[,,18:80],c(1,2),function(x){
    as.numeric(any(x>=1))
  })),ext=e)
  
  if(nrow(post_change_condensed)!=256){
    post_change_condensed <- disagg(post_change_condensed,fact=4)
  }
  polys <- (terra::as.polygons((patches((post_change_condensed),zeroAsNA=T)),values=T))
  
  tchange <- c(20,30,40,50,60)[c(3,1,5,2,4)]
  if(nrow(polys)>0){
    poly_results[[k]] <- list()
    
    poly_results[[k]]$polys <- polys
  }else{
    poly_results[[k]] <- list()
    poly_results[[k]]$polys <- NULL
  }
  
}


# ########## Plotting
# im <- rast((dat[,,1,80]))
# im2 <- rast((dat[,,2,80]))
# ##############
# tchange <- c(20,30,40,50,60)[c(3,1,5,2,4)]
# 
# 
# library(viridis)
# library(terra)
# # get polygons
# #tvec <- round(seq(1,length(tvals),length.out=length(tvals)/tchunk))
# truth_mask_condensed <- apply(truth_mask,c(1,2),function(x){
#   any(x>=1)
# })
# annots_flat <- terra::as.polygons((patches(rast(truth_mask_condensed),zeroAsNA=T)),values=T)
# annots_flat$dates <- tchange
# col_annots="cyan"
# 
# postscript("plots/JDS_paper/simulation_gen.eps", width=6, height=3, onefile=TRUE, horizontal=FALSE)
# par(mfrow=c(1,2))
# plot((im),axes=F,legend=F,main="Signal 1",col=viridis(100))
# plot(annots_flat,add=T,border=col_annots,lwd=2)
# text(annots_flat,annots_flat$dates,adj=c(-0.5,-0.5),col=col_annots)
# 
# plot((im2),axes=F,legend=F,main="Signal 2",col=viridis(100))
# plot(annots_flat,add=T,border=col_annots,lwd=2)
# text(annots_flat,annots_flat$dates,adj=c(-0.5,-0.5),col=col_annots)
# dev.off()
# 
# 
# plot(im,col=viridis(100),axes=F,legend=F)
# plot( poly_results[[1]]$polys,add=T,col="red",axes=F)
# plot( poly_results[[2]]$polys,add=T,col="blue",axes=F)
# plot(annots_flat,border="cyan",add=T,lwd=2,axes=F)
# legend("topright",legend=c("true change","prob 1","prob2"),
#        fill=c("white",c("red","blue")),border=c("cyan",rep(1,2)))
# 
# postscript("plots/JDS_paper/simulation_summary_probs_subset.eps", width=3, height=3, onefile=TRUE, horizontal=FALSE)
# par(mar=c(0,0,0,0))
# plot(im,col=viridis(100),axes=F,legend=F,mar=c(0,0,0,0))
# plot( poly_results[[3]]$polys,add=T,col="darkgreen",lwd=2,axes=F)
# plot( poly_results[[4]]$polys,add=T,col="blue",lwd=1,axes=F)
# plot(annots_flat,border="cyan",add=T,lwd=2,axes=F)
# dev.off()
# 
# postscript("plots/JDS_paper/simulation_summary_probs_subset_legend.eps", width=3, height=1.5, onefile=TRUE, horizontal=FALSE)
# par(mar=c(0,0,0,0))
# plot.new()
# legend("topright",legend=c("true change","prob c=1 subset","prob c=2 subset"),
#        fill=c("white",c("darkgreen","blue")),border=c("cyan",rep(1,2)))
# dev.off()
# 
# count.cols <- magma(5)[2:5]
# plot(im,col=viridis(100),axes=F,legend=F)
# plot( poly_results[[5]]$polys,add=T,col=count.cols[1],axes=F)
# plot( poly_results[[6]]$polys,add=T,col=count.cols[2],axes=F)
# plot( poly_results[[7]]$polys,add=T,col=count.cols[3],axes=F)
# plot( poly_results[[8]]$polys,add=T,col=count.cols[4],axes=F)
# #plot( poly_results[[9]]$polys,add=T,col=count.cols[5])
# #plot( poly_results[[10]]$polys,add=T,col=count.cols[5])
# plot(annots_flat,add=T,border="cyan",lwd=2,axes=F)
# legend("topright",legend=c("true change","count 1","count 2",
#                            "count 3","count 4"),
#        fill=c("white",count.cols),border=c("cyan",rep(1,4)))
# 
# postscript("plots/JDS_paper/simulation_summary_count_subset.eps", width=3, height=3, onefile=TRUE, horizontal=FALSE)
# par(mar=c(0,0,0,0))
# count.cols <- magma(7)[2:7]
# plot(im,col=viridis(100),axes=F,legend=F,mar=c(0,0,0,0))
# plot( poly_results[[11]]$polys,add=T,col=count.cols[1],axes=F)
# plot( poly_results[[12]]$polys,add=T,col=count.cols[2],axes=F)
# plot( poly_results[[13]]$polys,add=T,col=count.cols[3],axes=F)
# plot( poly_results[[14]]$polys,add=T,col=count.cols[4],axes=F)
# #plot( poly_results[[15]]$polys,add=T,col=count.cols[5],axes=F)
# #plot( poly_results[[16]]$polys,add=T,col=count.cols[6],axes=F)
# plot(annots_flat,add=T,border="cyan",lwd=2,axes=F)
# dev.off()
# 
# postscript("plots/JDS_paper/simulation_summary_count_subset_legend.eps", width=3, height=1.5, onefile=TRUE, horizontal=FALSE)
# par(mar=c(0,0,0,0))
# plot.new()
# legend("topright",legend=c("true change","count 1 subset","count 2 subset",
#                            "count 3 subset","count 4 subset"),
#        fill=c("white",count.cols),border=c("cyan",rep(1,6)))
# dev.off()
# 
# postscript("plots/JDS_paper/simulation_summary_classic.eps", width=3, height=3, onefile=TRUE, horizontal=FALSE)
# par(mar=c(0,0,0,0))
# plot(im,col=viridis(100),axes=F,legend=F,mar=c(0,0,0,0))
# plot( poly_results[[17]]$polys,add=T,col="red",border=F,lwd=2,axes=F)
# plot(annots_flat,border="cyan",add=T,lwd=2,axes=F)
# dev.off()
# 
# postscript("plots/JDS_paper/simulation_summary_classic_legend.eps", width=3, height=1.5, onefile=TRUE, horizontal=FALSE)
# par(mar=c(0,0,0,0))
# plot.new()
# legend("topright",legend=c("true change","roboBayes"),
#        fill=c("white",c("red")),border=c("cyan",rep(1,1)))
# dev.off()
# 
# 


